/* md5_2.c  -  MD5 CheckSum Manipulation
 *      Copyright (C) 2002 fkt <fkt@funfatal.org>
 *
 * This file is part of ProSum.
 *
 * ProSum is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * ProSum is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 */

/* Many code ripped from textutils package md5sum */
#include "md5_2.h"

#define STREQ(s1, s2) ((strcmp (s1, s2) == 0))

#if O_BINARY
# define OPENOPTS(BINARY) ((BINARY) != 0 ? TEXT1TO1 : TEXTCNVT)  
# define TEXT1TO1 "rb"
# define TEXTCNVT "r"
# else
#  if UNIX || __UNIX__ || unix || __unix__ || _POSIX_VERSION
#   define OPENOPTS(BINARY) "r"
#  else
    /* The following line is intended to evoke an error.
       Using #error is not portable enough.  */
    "Cannot determine system type."
#  endif
# endif

static int
md5_file (const char *filename, int binary, unsigned char *md5_result) {
  FILE *fp;
  int err;

  /* OPENOPTS is a macro.  It varies with the system.
     Some systems distinguish between internal and
     external text representations.  */

  fp = fopen (filename, OPENOPTS (binary));

  if (fp == NULL) return ERROR;

  err = md5_stream (fp, md5_result);
  if (err)
    {
      fclose (fp);
      return ERROR;
    }

  if (fclose (fp) == EOF) return ERROR;

  return SUCCESS;
}

int do_md5sum(file_t *f_st) {
  unsigned char md5buffer[16];
  int fail;

#if O_BINARY
  /* Binary is default on MSDOS, so the actual file contents
     are used in computation.  */
  int binary = 1;
#else
  /* Text is default of the Plumb/Lankester format.  */
  int binary = 0;
#endif

  fail = md5_file (f_st->filename, binary, md5buffer);

  if (fail==SUCCESS) {
    memcpy(f_st->f_md5,md5buffer,16);
    f_st->f_md5[16]='\0';
  }
  else return ERROR;

  return SUCCESS;
}

int check_md5sum(file_t *f_st, file_t *f_cmp) {
  size_t i;

  if (strncmp(f_st->f_md5,f_cmp->f_md5,16)) return ERROR;
  else return SUCCESS;
}
