/* snd_tcp_serv.c  -  TCP Server For Bastion Host
 *      Copyright (C) 2002 fkt <fkt@funfatal.org>
 *
 * This file is part of ProSum.
 *
 * ProSum is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * ProSum is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 */

#include "snd_tcp_serv.h"

int check_alias(char *alias) {
  DIR *cur_dir;
  struct dirent *f;
  int ret=ERROR;

  cur_dir=opendir("./");

  f=readdir(cur_dir);

  while (f != NULL) {
    if (strstr(f->d_name,alias)) {
      ret=SUCCESS;
      break;
    }

    f=readdir(cur_dir);
  }

  closedir(cur_dir);
  return ret;
}

int tiene_subdirs(char *ptr) {
  size_t i;

  for(i=0; i < strlen(ptr); i++)
    if (ptr[i]=='/') return 1;

  return 0;
}

RETSIGTYPE sigchld_handler(int signum) {
  while(wait(NULL) > 0);
}

RETSIGTYPE ctrlc_handler(int signum) {
  syslog(LOG_DAEMON | LOG_NOTICE,_("SIGINT Cached!, Terminating Server..."));
  exit(SUCCESS);
}

RETSIGTYPE sigterm_handler(int signum) {
  syslog(LOG_DAEMON | LOG_NOTICE,_("SIGTERM Cached!, Terminating Server..."));
  exit(SUCCESS);
}

int main(void) {
  char buf[MAX_BUF], cmd[MAX_CMD], tmp[MAX_CMD], *ptr_cmd, *ultimo_recv,
       *alias, alias_buf[MAX_ALIAS_SZ];
  struct sockaddr_in local_addr, remote_addr;
  int servfd, clientfd, tam_rmtaddr, fd, resto, code, l_ptr;
  off_t file_size;
  mode_t mode;
  unsigned int i, cont_recv, j=0, subdirs_ctrl=0, duplicate=0;
  struct stat status;
  struct sigaction sa;
  ssize_t c;
  struct utimbuf sync_times;

#ifdef HAVE_NLS
  //Ajusto el locale
  setlocale(LC_TIME,""); 
  if (setlocale(LC_MESSAGES,"")==NULL) {
    syslog(LOG_DAEMON | LOG_NOTICE,_("Unable To Set Locale"));
    FATAL("setlocale()");
  }

  bindtextdomain(PACKAGE, LOCALE_DIR);
  textdomain(PACKAGE);
#endif

  //Inicializacion para loguear en el syslog
  openlog(_("ProSum Server"),LOG_NDELAY,LOG_DAEMON);

  //Para no crear zombies
  sa.sa_handler=sigchld_handler; // Eliminar procesos muertos
  sigemptyset(&sa.sa_mask);
  sa.sa_flags=SA_RESTART;
  if (sigaction(SIGCHLD, &sa, NULL) == -1)
    syslog(LOG_DAEMON | LOG_NOTICE,_("Error executing sigaction()"));

  signal(SIGINT,ctrlc_handler);
  signal(SIGTERM,sigterm_handler);

  if (!geteuid())
    syslog(LOG_DAEMON | LOG_NOTICE,_("Warning!! You Are Running Server With Root UID"));

  if ((servfd=socket(AF_INET,SOCK_STREAM,0))==-1) {
    syslog(LOG_DAEMON | LOG_NOTICE,_("Unable To Create Socket"));
    FATAL("socket()");
  }

  local_addr.sin_family=AF_INET;
#ifdef WORDS_BIGENDIAN
  local_addr.sin_port=TCPSERV_PORT;
#else
  local_addr.sin_port=htons(TCPSERV_PORT);
#endif
  local_addr.sin_addr.s_addr=INADDR_ANY;

  if (bind(servfd,(struct sockaddr *)&local_addr, sizeof(struct sockaddr))==-1) {
    syslog(LOG_DAEMON | LOG_NOTICE,_("Unable To Bind Socket"));
    FATAL("bind()");
  }

  if (listen(servfd,TCPSERV_BACKLOG)==-1) {
    syslog(LOG_DAEMON | LOG_NOTICE,_("Unable To Listen"));
    FATAL("listen()");
  }

  syslog(LOG_DAEMON | LOG_NOTICE,_("Server Running, PID: %d"), getpid());

  while (1) {
    memset(cmd,0x0,sizeof(cmd));
    memset(buf,0x0,sizeof(buf));
    memset(tmp,0x0,sizeof(tmp));
    subdirs_ctrl=0;
    j=0;
    tam_rmtaddr=sizeof(remote_addr);

    if((clientfd=accept(servfd,(struct sockaddr *)&remote_addr,&tam_rmtaddr))==-1)
      syslog(LOG_DAEMON | LOG_NOTICE,_("Unable To Accept Client"));

    if (!fork()) {
      syslog(LOG_DAEMON | LOG_NOTICE,_("Detected Conection From %s"), inet_ntoa(remote_addr.sin_addr));
      read(clientfd,buf,sizeof(buf));

      if (strncmp(buf,"ProSum_0.28",12)) {
        syslog(LOG_DAEMON | LOG_NOTICE,_("Wrong Answer From %s, Disconnecting Him"), inet_ntoa(remote_addr.sin_addr));
        write(clientfd,&error,sizeof(int));
        close(clientfd);
        exit(ERROR);
      }
      syslog(LOG_DAEMON | LOG_NOTICE,_("Client Connected From %s"), inet_ntoa(remote_addr.sin_addr));
      write(clientfd,&ok,sizeof(int));

      memset(buf,0x0,sizeof(buf));
      read(clientfd,buf,sizeof(buf));

      //Pillamos el comando, pero antes miramos si es mas grande q MAX_CMD
      if (strlen(buf) > MAX_CMD) {
        syslog(LOG_DAEMON | LOG_NOTICE,_("%s Has Attemped To Overflow Us, Disconnecting Him"), inet_ntoa(remote_addr.sin_addr));
        write(clientfd,&overflow,sizeof(int));
        close(clientfd);
        exit(ERROR);
      }

      strncpy(cmd,buf,MAX_CMD);

      //Quito el \r\n del final de la cadena
      cmd[strlen(cmd)-2]='\0';

      //Miramos si el comando es setup, put o get
      if (strstr(cmd,"put")) {
        write(clientfd,&ok,sizeof(int));
        ptr_cmd=cmd+4; //Salto el comando y el espacio

        //Recibimos el alias y miramos si existe
        read(clientfd,alias_buf,sizeof(alias_buf));
        if (check_alias(alias_buf)==ERROR) {
          write(clientfd,&error,sizeof(int));
          close(clientfd);
          exit(ERROR);
        } else write(clientfd,&ok,sizeof(int));

        chdir(alias_buf);

        //Quito la ruta absoluta (si la tiene)
        if ((*ptr_cmd=='.') && ((*(ptr_cmd+1)=='/') || (*(ptr_cmd+1)=='.')))
          ptr_cmd++;
        while (*ptr_cmd == '/') ptr_cmd++;

        l_ptr=strlen(ptr_cmd);
        //Le quitamos todos los subdirs que tenga, para cogerlo del raiz
        while (tiene_subdirs(ptr_cmd) && j < l_ptr) {
          subdirs_ctrl=1;
          for (; *ptr_cmd != '/' ; j++, ptr_cmd++)
            ;;

          //Saltamos el /
          ptr_cmd++;
          j++;

          strcpy(tmp,ptr_cmd);
        }

        if (subdirs_ctrl)
          ptr_cmd=tmp;

        //Recibo el modo del fichero
        read(clientfd,&mode,sizeof(mode_t));

        fd=open(ptr_cmd, O_WRONLY | O_TRUNC | O_CREAT, mode);
        //Le cambiamos el modo, pq si el archivo ya existia no se cambia el modo solo
        fchmod(fd,mode);

        //Recibo el tamao del fichero y ajusto el contador del bucle para
        //recibir el fichero
        read(clientfd,&file_size,sizeof(off_t));
        cont_recv=file_size/MAX_BUF;

        //Recibo y escribo el fichero
        for (i=0; i < cont_recv ; i++) {
          memset(buf,0x0,sizeof(buf));
          read(clientfd,buf,sizeof(buf));
          write(fd,buf,sizeof(buf));
        }

        //Recibo y escribo lo que falte del fichero si no es multiplo de 1024
        resto=file_size%MAX_BUF;
        ultimo_recv=(char *)malloc(sizeof(char)*resto);
        read(clientfd,ultimo_recv,resto);
        write(fd,ultimo_recv,resto);

        free(ultimo_recv);
        close(fd);

        //Compruebo que el tamao del archivo que he recibido esta coincide
        //con el original
        stat(ptr_cmd,&status);

        if (status.st_size==file_size) {
          syslog(LOG_DAEMON | LOG_NOTICE,_("%s Recived From %s"), ptr_cmd, inet_ntoa(remote_addr.sin_addr));
          write(clientfd,&ok,sizeof(int));
        } else {
          syslog(LOG_DAEMON | LOG_NOTICE,_("Fail To Recive %s From %s"), ptr_cmd, inet_ntoa(remote_addr.sin_addr));
          write(clientfd,&error,sizeof(int));
        }

      } else if (strstr(cmd,"get")) {
        write(clientfd,&ok,sizeof(int));
        ptr_cmd=cmd+4;
        j=4;

        //Recibimos el alias y miramos si existe
        read(clientfd,alias_buf,sizeof(alias_buf));
        if (check_alias(alias_buf)==ERROR) {
          write(clientfd,&error,sizeof(int));
          close(clientfd);
          exit(ERROR);
        } else write(clientfd,&ok,sizeof(int));

        chdir(alias_buf);

        //Quito la ruta absoluta (si la tiene)
        if ((*ptr_cmd=='.') && ((*(ptr_cmd+1)=='/') || (*(ptr_cmd+1)=='.')))
          { ptr_cmd++; j++; }
        while (*ptr_cmd == '/') { ptr_cmd++; j++; }

        l_ptr=strlen(ptr_cmd);
        //Le quitamos todos los subdirs que tenga, para cogerlo del raiz
        while (tiene_subdirs(ptr_cmd) && j < l_ptr) {
          subdirs_ctrl=1;
          for (; *ptr_cmd != '/' ; j++, ptr_cmd++)
            ;;

          //Saltamos el /
          ptr_cmd++;
          j++;

          strcpy(tmp,ptr_cmd);
        }

        if (subdirs_ctrl)
          ptr_cmd=tmp;

        if ((fd=open(ptr_cmd,O_RDONLY))==-1) { //Si el archivo no se puede abrir
          write(clientfd,&nonext,sizeof(int));
          close(clientfd);
          exit(ERROR);
        } else write(clientfd,&ok,sizeof(int));

        //Mando el modo y tamao del archivo
        stat(ptr_cmd,&status);
        write(clientfd,&status.st_mode,sizeof(mode_t));
        write(clientfd,&status.st_size,sizeof(off_t));

        //Envio el archivo
        memset(buf,0x0,sizeof(buf));
        c=read(fd,buf,sizeof(buf));
        while (c != 0) {
          write(clientfd,buf,sizeof(buf));

          memset(buf,0x0,sizeof(buf));
          c=read(fd,buf,sizeof(buf));
        }

        close(fd);

        //Comprobamos si se ha recibido bien el archivo
        read(clientfd,&code,sizeof(int));
        if (code==ok)
          syslog(LOG_DAEMON | LOG_NOTICE,_("%s Sended To %s"), ptr_cmd, inet_ntoa(remote_addr.sin_addr));
        else
          syslog(LOG_DAEMON | LOG_NOTICE,_("Fail To Send %s To %s"), ptr_cmd, inet_ntoa(remote_addr.sin_addr));

      } else if (strstr(cmd,"setup")) {
        write(clientfd,&ok,sizeof(int));
        alias=cmd+6;

        if (check_alias(alias)==SUCCESS) {
          write(clientfd,&alias_dup,sizeof(int));
          close(clientfd);
          exit(ERROR);
        }

        if (mkdir(alias,S_IWUSR | S_IRUSR | S_IXUSR)==-1) {
          write(clientfd,&error,sizeof(int));
          close(clientfd);
          exit(ERROR);
        } else write(clientfd,&ok,sizeof(int));

        syslog(LOG_DAEMON | LOG_NOTICE,_("Alias Of %s Setup Recived From %s"), alias, inet_ntoa(remote_addr.sin_addr));

      } else { //FIN comprobacion del comando
        write(clientfd,&error,sizeof(int));
        syslog(LOG_DAEMON | LOG_NOTICE,_("Incorrect Command From %s"), inet_ntoa(remote_addr.sin_addr));
        close(clientfd);
        exit(ERROR);
      }

      close(clientfd);
      memset(cmd,0x0,sizeof(cmd));
      exit(SUCCESS);
    } else close(clientfd);
  }

  closelog();
  close(servfd);

  return 0;
}
